/*
 *  A simple Game Launcher
 *
 *  ...or is it?
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <iocs.h>
#include <string.h>
#include "misc.h"
#include "Joypad.h"

#define TEXTMODE_WIDTH				96
#define TEXTMODE_HEIGHT				31

#define COLOUR_BLACK                 4
#define COLOUR_BLUE                  5
#define COLOUR_YELLOW                6
#define COLOUR_WHITE                 7

#define MAX_GAMES					1000

#define UPPER(c) ((c >= 'a' && c <= 'z') ? (c - 'a') + 'A': c)

typedef struct GameInfo GameInfo;

struct GameInfo
{
	short length;
	char name[50];
	char foldername[100];
} __attribute__ ((aligned (2)));

void Print(int x, int y, void *text, int color)
{
    _dos_c_locate(x,y);
    _dos_c_color(color);
    _dos_c_print(text);
}

void Putc(int x, int y, char chr, int color)
{
    _dos_c_locate(x,y);
    _dos_c_color(color);
    _dos_c_putc(chr);
}

void CLS()
{
    _dos_c_cls_al();
}

void Cleanup()
{
	CLS();
    _dos_c_color(3);
    _dos_c_locate(0, 26);
    _dos_c_curon();
}

void *heap = null;
int games = 0;
int longest = 0;

GameInfo *ginfo[MAX_GAMES];

void InsertGame(GameInfo *game)
{
	bool inserted = false;
	int i;
	GameInfo *tmp;

    for (i = 0; i < MAX_GAMES; i++)
	{
		if (ginfo[i] == NULL)
		{
			ginfo[i] = game;
			return;
		}

		if (inserted)
		{
			tmp = ginfo[i];
			ginfo[i] = game;
			game = tmp;
			continue;
		}

		if (strcoll(ginfo[i]->name, game->name) > 0)
		{
			inserted = true;

			tmp = ginfo[i];
			ginfo[i] = game;
			game = tmp;
		}
	}
}

void ScanFolder(int driveno, char *folder)
{
	bool initalized = false, end;
	int i, j, ret;
	struct dos_exfilbuf fbuffer;
	GameInfo *g;

	ret = _dos_chgdrv(driveno);
	ret = _dos_chdir(folder);

	while(ret >= 0)
	{
		if (!initalized)
		{
			initalized = true;
			ret = _dos_exfiles(&fbuffer, "*", 0x10);
		}
		else
			ret = _dos_exnfiles(&fbuffer);

		if (ret < 0)
			continue;

		if (fbuffer.name[0] == '.')
			continue;

		g = (GameInfo *)heap+(games*sizeof(GameInfo));

		for (i = 0, j = 0; i < sizeof(fbuffer.name); i++, j++)
		{
			if (i > 0 && fbuffer.name[i-1] != '_' && ((fbuffer.name[i] >= 'A' && fbuffer.name[i] <= 'Z' && (fbuffer.name[i-1] < 'A' || fbuffer.name[i-1] > 'Z')) || (fbuffer.name[i] >= '0' && fbuffer.name[i] <= '9' && (fbuffer.name[i-1] < '0' || fbuffer.name[i-1] > '9') )))
			{
				g->name[j] = ' ';
				j++;
			}

			if (fbuffer.name[i] != '_')
				g->name[j] = fbuffer.name[i];
			else
				g->name[j] = ' ';

			if (fbuffer.name[i] == '\0')
			{
					g->length = j;
					if (j > longest)
						longest = j;
					break;
			}
		}

		sprintf(g->foldername,"%s\\%s", folder, fbuffer.name);
		InsertGame(g);
		games++;
	}

	for (i = 0; i < games; i++)
	{
		end = false;

		for (j = 0; j < longest; j++)
		{
			if (ginfo[i]->name[j] == '\0')
				end = true;

			if (end)
				ginfo[i]->name[j] = ' ';
		}
		ginfo[i]->name[j] = '\0';
	}

    //free(fbuffer);
}

int main (int argc, char **argv)
{
	bool initalized = false, listmode = true, update = true;
	unsigned char text[TEXTMODE_WIDTH];
	unsigned char letter, blank[50], c;
	int i, j, ret, selection = 0, listheight, colour, offset;
	int fin, fout;
	struct dos_filbuf fbuffer;

	if (argc > 1 && argv[1][0] == '1')
	{
		PollJoypad(0, &joypad1);

		if (!joypad1.a && !joypad1.b)
			return 0;
	}

	_iocs_b_super(0);

	//fbuffer = malloc(sizeof(struct dos_filbuf));
	//heap = malloc(sizeof(GameInfo) * MAX_GAMES);
	heap = _dos_malloc(sizeof(GameInfo) * MAX_GAMES);

	if (heap >= (void*)0x81000000)
	{
		printf("Memory allocation error!\n");
		return 0;
	}

    CLS();
    _dos_c_curoff();

	for (i = 0; i < MAX_GAMES; i++)
		ginfo[i] = NULL;

	//ret = _dos_chgdrv(0);
	ret = _dos_chdir("\\");

	while(ret >= 0)
	{
		if (!initalized)
		{
			initalized = true;
			ret = _dos_files(&fbuffer, "GAMES*", 0x10);
		}
		else
			ret = _dos_nfiles(&fbuffer);

		if (ret < 0)
			continue;

		ScanFolder(fbuffer.driveno, fbuffer.name);
		_dos_chdir("\\");
	}

	for (i = 0; i < longest-1; i++)
		blank[i] = ' ';
	blank[i] = '\0';

	for (i = 0; i < sizeof(text)-1; i++)
		text[i] = ' ';
	text[i] = '\0';

	text[(TEXTMODE_WIDTH>>1)-(longest>>1)-2] = 0x7C;
	text[(TEXTMODE_WIDTH>>1)+(longest>>1)+2] = 0x7C;

	for (i = 0; i < TEXTMODE_HEIGHT; i++)
		Print(0, i, text, COLOUR_YELLOW);

	listheight = (games < TEXTMODE_HEIGHT) ? games : TEXTMODE_HEIGHT;

	text[2] = 0x7B;
	text[4] = 0x7D;
	text[5] = '\0';

	if (games > 0)
	{
		letter = UPPER(ginfo[0]->name[0]);
		if (letter < 'A' || letter > 'Z')
			letter = 'A';
	}
	else
	{
		printf("No 'GAMES*' folders found on current drive!\n");
		goto END;
	}

	do
	{
		//wait_for_vblank();
		//CLS();
		PollJoypad(0, &joypad1);

		if (joypad1.a)
		{
            selection = -1;
            break;
		}

		if (joypad1.down)
		{
			update = true;
			if (listmode)
			{
				selection++;
				if (selection >= games)
					selection = games-1;
			}
			else
			{
                letter++;
                if (letter > 'Z')
					letter = 'Z';

				for (i = 0; i < games; i++)
					if (UPPER(ginfo[i]->name[0]) == letter)
					{
						selection = i;
						break;
					}
			}
		}

		if (joypad1.up)
		{
			update = true;
			if (listmode)
			{
				selection--;
				if (selection < 0)
					selection = 0;
			}
			else
			{
                letter--;
                if (letter < 'A')
					letter = 'A';

				for (i = 0; i < games; i++)
					if (UPPER(ginfo[i]->name[0]) == letter)
					{
						selection = i;
						break;
					}
			}
		}

		if (joypad1.left && listmode)
		{
			update = true;
			listmode = false;

			letter = UPPER(ginfo[selection]->name[0]);
		}

		if (joypad1.right && !listmode)
		{
			update = true;
			listmode = true;

			//selection = 0;
			for (i = 0; i < games; i++)
				if (UPPER(ginfo[i]->name[0]) == letter)
				{
					selection = i;
					break;
				}
		}

		if (update)
		{
			offset = 0;

			if (selection >= TEXTMODE_HEIGHT>>1)
			{
				offset = selection - (TEXTMODE_HEIGHT>>1);
			}
			//offset

			for (i = 0; i < listheight; i++)
			{
				if (listmode && i+offset == selection)
				{
					letter = UPPER(ginfo[i+offset]->name[0]);
					colour = COLOUR_YELLOW;
				}
				else colour = COLOUR_WHITE;

				if (i+offset >= games)
					Print((TEXTMODE_WIDTH>>1)-(longest>>1)/*(ginfo[i+offset]->length>>1)*/, i, blank, colour);
					//continue;
				else
					Print((TEXTMODE_WIDTH>>1)-(longest>>1)/*(ginfo[i+offset]->length>>1)*/, i, ginfo[i+offset]->name, colour);
			}

			for (i = 0, c = 'A'; c <= 'Z'; c++, i++)
			{
				if (c == letter)
				{
					if (listmode)
						colour = COLOUR_BLUE;
					else
						colour = COLOUR_YELLOW;
				}
				else
					colour = COLOUR_WHITE;

				text[3] = c;
				Print(2, (TEXTMODE_HEIGHT >> 1)-13+i,text, colour);

				//Putc(3, (TEXTMODE_HEIGHT >> 1)-13+i,c, colour);
			}

			update = false;
		}
		//wait_for_vdisp();
	}
	while(!(joypad1.b && listmode));

	_dos_chdir("\\");
	fout = _dos_create("!Start.bat",0);

	if (selection >= 0)
	{
		_dos_chdir(ginfo[selection]->foldername);
		fin = _dos_open("!Start.bat", 0);

		ret = sprintf(text, "CD %s\r\n", ginfo[selection]->foldername);
		_dos_write(fout,text,ret);

		do
		{
			ret = _dos_read(fin, text, sizeof(text));

			if (ret > 0)
				_dos_write(fout,text,ret);

		}
		while(ret > 0);

		_dos_close(fin);
	}
	_dos_close(fout);

	END:
    _dos_mfree(heap);
    //free(fbuffer);
    Cleanup();
    return 0;
}
